/***************************************************************************
 *            stp.c
 *
 *  Thu Jun 16 01:58:55 2005
 *  Copyright  2005  Carbone Francesco
 *  trini4mail@yahoo.it
 *	franci.carbone@gmail.com
 ****************************************************************************

 ****************************************************************************
 *	This code implement functions libraries needed to implement the 
 * 	Spanning Tree Algorithm & Protocol
 *	The Spanning Tree Algorithm and its associated Bridge Protocol 
 *	operate to support, preserve, and mantain the quality of the 
 *	MAC Bridges Service in all its aspect. In order to perform this 
 *	functions, the algorithm meets the following requirements :
 *
 *	a) 	It will configure the active topology og Bridged LAN of arbitrary
 *		topology into a single spanning tree, such that there is at most
 *		one data route between any two end stations, eliminating data loops.
 *
 *	b)	It will provide for fault tolerance by automatic reconfiguration
 *		of the spanning tree topology as a result of Bridge failure or a 
 * 		breakdown in data path.
 *
 *	c)	The entire active topologyy will stabilize in any Bridged LAN
 *		within a short known bounded time interval.
 *
 *	d)	The active topology will be predictable and reproducible, and may 
 *		be selected  by management of the parameters of the algotithm.
 *	
 *	e)	It will operate transparently to the end stations.
 *
 *	f)	The communication bandwith consumed by the Bridges in enstablishing
 *		and mantain the spanning tree on a particular lan will be a small 
 *		percentage of the total available bandwith and independent of the 
 * 		total traffic supported by the Bridged LAN regardles of the total 
 *		number of Bridges or LANs.
 *
 *	g)	The memory requirements associated with each Bridge Port are 
 *		independent of the number of Bridges and LANs in the Bridged LAN.
 *
 *	h)	Bridge do not have to be individually configured before being added 
 *	 	to the Bridged LAN, other than having their MAC Addresses assigned 
 *		through normal procedures.
 ****************************************************************************
 */
 
 #include 	"stp.h"
 
 
/* Forward Declarations */
static void 	stp_topology_change_detection(struct stp_bridgedata *);
static int		stp_root_bridge(struct stp_bridgedata);

/* 	******FUNCTIONS USED ONLY FOR TESTING *****************/

/*
 *	This section is implemented out from 	*
 * 	here, because testing 					*
*/
extern void stp_send_config_bpdu(struct stp_bridgedata *, uint16_t, struct  stp_configbpdu);	// TO BE IMPLEMENTED
extern void stp_send_tcn_bpdu(struct stp_bridgedata *, uint16_t , struct stp_configbpdu); 		// TO BE IMPLEMENTED

/* ****** ONLY FOR TESTING ********************************
 * 	Create Bridge Data Structure receiving its configuration 
 *	from a file in the form :	
 *
 *	Bridge-Address Bridge-priority maxAge helloTime forwardDelayTime numOfPorts	
 *	Port-number LAN-connected Port-priority pathCost 							
 *		'			'			'		'								
 *		'			'			'		'								
 *	Port-number LAN-connected Port-priority pathCost					
 */

/* Allocate dinamic memory */
void *
myalloc(size_t size)
	{
	void *space;
	space = (void *) malloc(size);
	return (space);
	}

/* 
 *	Use input configuration file to store  Bridge 
 *	and Port parameters
 */	
struct stp_bridgedata *
stp_create_bridge(const char *filename)	
{	
	uint16_t 	p;
	uint16_t 	i=1;
	FILE 		*fp;
	uint16_t 	pNum = One;

	struct stp_bridgedata *bridgeData;

	if ((fp=fopen(filename,"r")) <= 0) 
		return NULL;

	
	
	/* 	
	 *	This part is used to test the correct form of configuration file 
	 *	and information stored in
	 */
	
	if ( (bridgeData = (struct stp_bridgedata *) myalloc(sizeof(struct stp_bridgedata))) == NULL ) 
		return NULL;
	/*
	 *	Informations about Bridge parameters are loaded from file, 
	 *	one line is loaded
	 */
	if (fscanf(fp,"%lld %hd %hd %hd %hd %hd",&bridgeData->bd_bridgeid,&p,&bridgeData->bd_bridgemaxage,&bridgeData->bd_bridgehellotime,&bridgeData->bd_bridgeforwarddelaytime,&bridgeData->bd_numofports) !=6)
		{	
			free(bridgeData);
			return NULL;
		}
		
	/*
	 *	Bridge Address, Bridge Priority, Bridge Max Age, Bridge Hello Time, 
	 *	Bridge Forward Delay and number of ports are displayed on terminal
 	*	
 	*/
	printf("Bridge A %li Bridge P %i B_maxAge %i B_helloTime %i  B_Forward_D %i B_numports %i\n",\
		(long int)bridgeData->bd_bridgeid,(int)p,(int)bridgeData->bd_bridgemaxage,\
		(int)bridgeData->bd_bridgehellotime,(int)bridgeData->bd_bridgeforwarddelaytime,\
		(int)bridgeData->bd_numofports);
	stp_set_priority_bridgeid(&bridgeData->bd_bridgeid,p);
	bridgeData->bd_holdtime = HOLD_TIME;
	bridgeData->bd_topologychangetime = TOPOLOGY_CHANGE_TIME;
	if ((bridgeData->bd_ports = (struct stp_portdata **) myalloc(((bridgeData->bd_numofports)*sizeof(struct stp_portdata *))+1)) == NULL) 
		{	
			free(bridgeData);
			return  NULL;
		}
	if ((bridgeData->bd_ports[0] = (struct stp_portdata *) myalloc (sizeof(struct stp_portdata))) == NULL)
		{	
			free(bridgeData);
			return NULL;
		}
	bridgeData->bd_ports[0]->pd_designatedroot = bridgeData->bd_bridgeid ;
	bridgeData->bd_ports[0]->pd_pathcost = Zero;
	bridgeData->bd_ports[0]->pd_designatedcost = Zero;
	
	while ( pNum <= (bridgeData->bd_numofports) )
		{	
			if ((bridgeData->bd_ports[pNum] = (struct stp_portdata *) myalloc (sizeof(struct stp_portdata))) == NULL)
				{	i=1;
					for(i=One;i<pNum;i++)
						{ free(bridgeData->bd_ports[i]); }
					free(bridgeData);
					return NULL;
				}
			/*
 		 	 *	Port parameters are loaded from file, 
 			 *	one line at a time for each port
			 */
			if (fscanf(fp,"%hd %hd %hd %d",&bridgeData->bd_ports[pNum]->pd_portid,\
					&bridgeData->bd_ports[pNum]->pd_index,&p,&bridgeData->bd_ports[pNum]->pd_pathcost) != 4)
				{   
					i=1;
					for(i=1;i<=pNum;i++)
						{ free(bridgeData->bd_ports[i]); }
					free(bridgeData);
					printf("Configuration file bridge.conf%lld error\n",\
						stp_get_address_bridgeid(bridgeData->bd_bridgeid));
					return NULL;
				}
			
			stp_set_priority_portid(&bridgeData->bd_ports[pNum]->pd_portid,p);
			/*
 			 *	Port number, Cost, Port number and port priority are displayed 
			 *	on terminal
			 */
			
			printf("Port %d Cost %d PortNum %d PortPrior %d\n",(int)bridgeData->bd_ports[pNum]->pd_portid,\
				(int)bridgeData->bd_ports[pNum]->pd_pathcost,\
				(int)stp_get_portno_portid(bridgeData->bd_ports[pNum]->pd_portid),\
				(int)stp_get_priority_portid(bridgeData->bd_ports[pNum]->pd_portid));
			pNum++;
		}
	return (bridgeData);
}	

/* Free Bridge allocated dinamic memory */
void
stp_free(struct stp_bridgedata *bridgeData)
{
	uint16_t numx = 0;

	for(numx=0; numx<=bridgeData->bd_numofports; numx++)
		free(bridgeData->bd_ports[numx]);
	free(bridgeData->bd_ports);
	free(bridgeData);
}
	

// END OF TESTING PART



/*
 *******************************************************************************
 ********* FUNCTIONS LIBRIRIES TO IMPLEMENT SPANNING TREE PROTOCOL *************
 *******************************************************************************
 */

/* 
 *	Transmit a configuration Bpdu trought the port specified, if the hold timer 
 *	is not active
 */
static void 	
stp_transmit_configuration_bpdu(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	struct stp_configbpdu	configBpdu;
	struct stp_portdata 	*port = bridgeData->bd_ports[portNo];
	if (stp_is_timer_active(&port->pd_timerhold))	
		{	
			port->pd_configpending = True;
		}
	else 	
		{	
			configBpdu.cbpdu_protoid = BPDU_PROTOCOL_IDENTIFIER;
			configBpdu.cbpdu_protover = BPDU_PROTOCOL_VERSION;
			configBpdu.cbpdu_bpdutype = CONFIG_BPDU_TYPE;
			configBpdu.cbpdu_designatedroot = bridgeData->bd_designatedroot;
			configBpdu.cbpdu_rootpathcost = bridgeData->bd_rootpathcost;
			configBpdu.cbpdu_bridgeid = bridgeData->bd_bridgeid;
			configBpdu.cbpdu_portid  = port->pd_portid;
			if (stp_root_bridge(*bridgeData))
				{	configBpdu.cbpdu_messageage = Zero; }
			else	
				{ 
				configBpdu.cbpdu_messageage =((((bridgeData->bd_maxage)-(stp_get_timer_value(bridgeData->bd_ports[stp_get_portno_portid(bridgeData->bd_rootport)]->pd_timermessageage)))+MESSAGE_AGE_INCREMENT)); 
				}
			configBpdu.cbpdu_maxage = bridgeData->bd_maxage ;
			configBpdu.cbpdu_hellotime = bridgeData->bd_hellotime ;
			configBpdu.cbpdu_forwarddelaytime = bridgeData->bd_forwarddelaytime;
			if (port->pd_topologychangeack) 	stp_set_tcack(&configBpdu.cbpdu_flags);
				else	stp_reset_tcack(&configBpdu.cbpdu_flags);
			if (bridgeData->bd_topologychange)	stp_set_tc(&configBpdu.cbpdu_flags);
				else stp_reset_tc(&configBpdu.cbpdu_flags);
			if (configBpdu.cbpdu_messageage < bridgeData->bd_maxage )	{	
				port->pd_topologychangeack = False;
				port->pd_configpending = False ;
				stp_send_config_bpdu(bridgeData,portNo,configBpdu); 				// Send a Bpdu Configuration message trought the port
				stp_start_timer(&port->pd_timerhold, bridgeData->bd_holdtime); 		// Restart hold timer for this port
				}
 		}
 		
}

/* 	Set port STATE :
	-disabled	0
	-blocking	1
	-listening	2
	-learning	3
	-forwarding	4
*/
static void 	
stp_set_port_state(struct stp_portdata *port, uint8_t state)
{	
	port->pd_state = state;
}

/*	Test if the Bridge is the Root */
static int 	
stp_root_bridge(struct stp_bridgedata bridgeData)
{
	return (bridgeData.bd_bridgeid == bridgeData.bd_designatedroot);
}

/* Test if the Bridge is the designated Bridge for the LAN where the port is attached */
static int 	
stp_designated_bridge(struct stp_bridgedata *bridgeData,uint16_t portNo)
{	
	return ( bridgeData->bd_ports[portNo]->pd_designatedbridge == bridgeData->bd_bridgeid );
}

/* Test if the port is the designated port */
int 	// STATIC "PLEASE SET THIS STATIC OUT OF TESTING ITS BETTER"
stp_designated_port(struct stp_bridgedata *bridgeData,uint16_t portNo)
{		
	return ( (bridgeData->bd_ports[portNo]->pd_designatedport == bridgeData->bd_ports[portNo]->pd_portid) && ( bridgeData->bd_ports[portNo]->pd_designatedbridge == bridgeData->bd_bridgeid ));
}

/* Test if the Port is root port */
int // STATIC	"PLEASE SET THIS STATIC OUT OF TESTING"	
stp_root_port(struct stp_bridgedata *bridgeData,uint16_t portNo)
{	
	return ( bridgeData->bd_rootport == bridgeData->bd_ports[portNo]->pd_portid );
}

/* 	
 *	To record , for a port, protocol parameters conveyed by a configuration 
 *	BPDU received on that port
 */
static void 	
stp_record_configuration_information(struct stp_bridgedata *bridgeData, uint16_t portNo, struct stp_configbpdu configBpdu)
{	struct stp_portdata *port = bridgeData->bd_ports[portNo];
	port->pd_designatedroot = configBpdu.cbpdu_designatedroot;
	port->pd_designatedcost = configBpdu.cbpdu_rootpathcost;
	port->pd_designatedbridge = configBpdu.cbpdu_bridgeid;
	port->pd_designatedport = configBpdu.cbpdu_portid ;
	stp_start_timer(&port->pd_timermessageage, (bridgeData->bd_maxage-configBpdu.cbpdu_messageage) );
}

/*	
 *	To update the Max Age , Hello Time , Forward Delay , and Topology Change 
 *	flag parameters to the last values received from the Root
 */
static void 
stp_record_configuration_timeout_values(struct stp_bridgedata *bridgeData, struct stp_configbpdu configBpdu)
{	
	bridgeData->bd_maxage = configBpdu.cbpdu_maxage;
	bridgeData->bd_hellotime = configBpdu.cbpdu_hellotime;
	bridgeData->bd_forwarddelaytime = configBpdu.cbpdu_forwarddelaytime;
	bridgeData->bd_topologychange = stp_is_tc_set(configBpdu.cbpdu_flags);
}

/*  
 *	To convey to Bridges attached to each LAN for which the Bridge is Designated
 *	Bridge knowledge of the Designated Root, Root Path Cost, Designated Bridge, 
 *	Designated Port, and the values of protocol timers
 */
static void 	
stp_configuration_bpdu_generation(struct stp_bridgedata *bridgeData)
{	
	uint16_t portNo = Zero;

	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{
			if (( stp_designated_port(bridgeData,portNo) && ( bridgeData->bd_ports[portNo]->pd_state != Disabled)))
				stp_transmit_configuration_bpdu(bridgeData,portNo);
		}
}
	
static void 	
stp_replay_to_configuration_bpdu(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	stp_transmit_configuration_bpdu(bridgeData,portNo);
}

/*  
 *	Transmit a Topology Change Notification BPDU trought root port, notifing 
 *	the bridge on the path towards the root port or that an extention of topology has
 *	been detected by the transmitting Bridge
 */	
static void 	
stp_transmit_tcn_bpdu(struct stp_bridgedata *bridgeData)
{
	// TEMP , ONLY FOR TESTING "PLESE REMOVE THIS TO USE FUNCTION OUT OF TESTING"
	struct stp_configbpdu 	tcnBpdu;
	uint16_t		portNo;
	tcnBpdu.cbpdu_protoid = BPDU_PROTOCOL_IDENTIFIER;
	tcnBpdu.cbpdu_protover = BPDU_PROTOCOL_VERSION;
	tcnBpdu.cbpdu_bpdutype = TCN_BPDU_TYPE;
	portNo = stp_get_portno_portid(bridgeData->bd_rootport);
	stp_send_config_bpdu(bridgeData,portNo,tcnBpdu);
	
/**** PLEASE UNCOMMENT THIS OUT OF TESTING AND CANCEL CODE ABOVE **********
	struct stp_tcnbpdu 	tcnBpdu;
	uint16_t		portNo;
	tcnBpdu.tcn_protoid = BPDU_PROTOCOL_IDENTIFIER;
	tcnBpdu.tcn_protover = BPDU_PROTOCOL_VERSION;
	tcnBpdu.tcn_bpdutype = TCN_BPDU_TYPE;
	portNo = stp_get_portno_portid(bridgeData->stpbd_rootport);
	stp_send_tcn_bpdu(bridgeData,portNo,tcnBpdu);
*/
}

	
/* 	
 *	Given that a Port is to be a Designated Port on the LAN to which it is attached,
 *	to assign appropriated values to those Port parameters that determine the 
 *	active topology of the Bridged LAN
 */
static void	
stp_become_designated_port(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	struct stp_portdata *port;
	port = bridgeData->bd_ports[portNo];
	port->pd_designatedroot = bridgeData->bd_designatedroot;
	port->pd_designatedcost = bridgeData->bd_rootpathcost;
	port->pd_designatedbridge = bridgeData->bd_bridgeid;
	port->pd_designatedport = port->pd_portid;
}

/* 
 *	To select the Designated Root and the Root Port, and to calculate the Root 
 * 	Path Cost for this Bridge
 */
static void	
stp_root_selection(struct stp_bridgedata *bridgeData)
{	
	uint16_t	portNo, rootPort;
	struct stp_portdata **bP;
	rootPort = Zero;
	bP = bridgeData->bd_ports;
	
	/* 	
	 *	Not necessary , but it's better secure in case programmers
	 *	forgot something
	*/	

	bP[rootPort]->pd_designatedroot = bridgeData->bd_bridgeid;
	bP[rootPort]->pd_pathcost = Zero;
	bP[rootPort]->pd_designatedcost = Zero;

	// end 

	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	
		if (	( 	(!(stp_designated_port(bridgeData,portNo) )) 
					&& 
					(bP[portNo]->pd_state != Disabled) 
					&&
					(bP[portNo]->pd_designatedroot < bridgeData->bd_bridgeid)
				)
				&&
				( 	(rootPort == Zero)
					||
					(bP[portNo]->pd_designatedroot < bP[rootPort]->pd_designatedroot )
					||
					( 	(bP[portNo]->pd_designatedroot == bP[rootPort]->pd_designatedroot) 
						&& 
						(	( 	(bP[portNo]->pd_designatedcost + bP[portNo]->pd_pathcost) 
								< 
								( bP[rootPort]->pd_designatedcost + bP[rootPort]->pd_pathcost) 
							)	
							|| 
							(	( 	( (bP[portNo]->pd_designatedcost + bP[portNo]->pd_pathcost) 
									== 
									( bP[rootPort]->pd_designatedcost + bP[rootPort]->pd_pathcost) 
									)
								)
								&& 
								(	(
									bP[portNo]->pd_designatedbridge
									< 
									bP[rootPort]->pd_designatedbridge
									)
									||  
									(	(
											bP[portNo]->pd_designatedbridge 
											== 
											bP[rootPort]->pd_designatedbridge
										) 
										&& 
										(	( 
											bP[portNo]->pd_designatedport 
											< 
											bP[rootPort]->pd_designatedport
											) 
											|| 
											(	( 	bP[portNo]->pd_designatedport
													== 
													bP[rootPort]->pd_designatedport
												) 
												&& 
												(	bP[portNo]->pd_portid 
													< 
													bP[rootPort]->pd_portid
												)
											)
										)
									)
								)
							)
						)
					)
				)
			)
			{	
				rootPort = portNo;
			}
			
		}
	bridgeData->bd_rootport = bP[rootPort]->pd_portid;
	if (rootPort == Zero)
		{	bridgeData->bd_designatedroot = bridgeData->bd_bridgeid;
			bridgeData->bd_rootpathcost = Zero;
		}
	else
		{	
			bridgeData->bd_designatedroot = bP[rootPort]->pd_designatedroot;
			bridgeData->bd_rootpathcost = bP[rootPort]->pd_designatedcost + bP[rootPort]->pd_pathcost;
		}
}

/*	
 *	To determine, for each Port, whether the Port should be the designated Port 
 *	for the Lan to which it is attached
*/
static void 	
stp_designated_port_selection(struct stp_bridgedata *bridgeData)
{	
	uint16_t	portNo = One;
	struct stp_portdata **bP = bridgeData->bd_ports;

	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	if (	stp_designated_port(bridgeData,portNo) 
					|| 
					(	bP[portNo]->pd_designatedroot != bridgeData->bd_designatedroot) 
					|| 
					( bridgeData->bd_rootpathcost < bP[portNo]->pd_designatedcost ) 
					|| 
					(	(bridgeData->bd_rootpathcost == bP[portNo]->pd_designatedcost) 
						&& 
						(	(bridgeData->bd_bridgeid < bP[portNo]->pd_designatedbridge)
							|| 
							(	(bridgeData->bd_bridgeid == bP[portNo]->pd_designatedbridge) 
								&& 
								(bP[portNo]->pd_portid < bP[portNo]->pd_designatedport)
							)
						)
					)
				)
			{	
				stp_become_designated_port(bridgeData,portNo);
			}
		}
}
	

/* To update configuration information held for the Bridge and Bridge Ports */
static void	
stp_configuration_update(struct stp_bridgedata *bridgeData)
{	
	stp_root_selection(bridgeData);
	stp_designated_port_selection(bridgeData);
}

/*	
 *	To permit a port to partecipate in frame relay, following a suitable interval
 *	to ensure that temporary loops in the Bridged LAN do not cause duplication of
 *	frames
 */	
static void 	
stp_make_forwarding(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	if (bridgeData->bd_ports[portNo]->pd_state == Blocking)
		{	bridgeData->bd_ports[portNo]->pd_state = Listening;
			stp_start_timer(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay, bridgeData->bd_forwarddelaytime);
		}
}

/*	To terminate the partecipation of a Port in frame relay */
static void 	
stp_make_blocking(struct stp_bridgedata *bridgeData,uint16_t portNo)
{	
	if (	(bridgeData->bd_ports[portNo]->pd_state != Disabled) 
			&& 
			(bridgeData->bd_ports[portNo]->pd_state != Blocking)
		)
			{	if 	(	
						(bridgeData->bd_ports[portNo]->pd_state == Forwarding) 
						|| 
						(bridgeData->bd_ports[portNo]->pd_state == Learning)
					)
					{	
						stp_topology_change_detection(bridgeData);
					}
				stp_set_port_state(bridgeData->bd_ports[portNo],Blocking);
				stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay);
			}
}

		
/*	To select the state of Bridge's Ports */
static void 	
stp_port_state_selection(struct stp_bridgedata *bridgeData)
{	
	uint16_t	portNo = One;
	
	struct stp_portdata **bP = bridgeData->bd_ports;

	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	if (stp_root_port(bridgeData,portNo))
				{	bP[portNo]->pd_configpending = False;
					bP[portNo]->pd_topologychangeack = False;
					stp_make_forwarding(bridgeData,portNo);
				}
			else	
				{	if (stp_designated_port(bridgeData,portNo))
						{	
							stp_stop_timer(&bP[portNo]->pd_timermessageage);
							stp_make_forwarding(bridgeData,portNo);
						}
					else
						{	bP[portNo]->pd_configpending = False;
							bP[portNo]->pd_topologychangeack = False;
							stp_make_blocking(bridgeData,portNo);
						}
				}
		}
}

/*	To record a topology change that has been detected by or notified to the Bridge */
static void 	
stp_topology_change_detection(struct stp_bridgedata *bridgeData)
{	
	if (stp_root_bridge(*bridgeData))
		{	bridgeData->bd_topologychange = True;
			stp_start_timer(&bridgeData->bd_timertopologychange,bridgeData->bd_topologychangetime);
		}
	else
		{
			if ( (bridgeData->bd_topologychangedetected) == False)
				{
					stp_transmit_tcn_bpdu(bridgeData);
					stp_start_timer(&bridgeData->bd_timertcn, bridgeData->bd_bridgehellotime);
				}
		}
	bridgeData->bd_topologychangedetected = True;
}

/* To terminate the transmission of Topology Change Notification BPDUs */
static void 	
stp_topology_change_ack(struct stp_bridgedata *bridgeData)
{	
	bridgeData->bd_topologychangedetected = False;
	stp_stop_timer(&bridgeData->bd_timertcn);
}

/* To acknowledge notification of a detected topology change by another stp_bridgedata */
static void 	
stp_ack_topology_change(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	bridgeData->bd_ports[portNo]->pd_topologychangeack = True;
	stp_transmit_configuration_bpdu(bridgeData,portNo);
}

/* To verify if a configuration Bpdu , supersedes information recorded  */
static int 	
stp_bpdu_info_compare(struct stp_bridgedata *bridgeData, uint16_t portNo, struct stp_configbpdu configBpdu)
{
	return 	( 	(configBpdu.cbpdu_designatedroot < bridgeData->bd_ports[portNo]->pd_designatedroot )
				|| 
				( 	(	configBpdu.cbpdu_designatedroot == bridgeData->bd_ports[portNo]->pd_designatedroot) 
					&&	
					( 	( configBpdu.cbpdu_rootpathcost < bridgeData->bd_ports[portNo]->pd_designatedcost )
						|| 
						( 	( configBpdu.cbpdu_rootpathcost == bridgeData->bd_ports[portNo]->pd_designatedcost )
							&& 
							( 	( configBpdu.cbpdu_bridgeid < bridgeData->bd_ports[portNo]->pd_designatedbridge)
								|| 
								( ( configBpdu.cbpdu_bridgeid == bridgeData->bd_ports[portNo]->pd_designatedbridge )
									&& 
									( 	( configBpdu.cbpdu_bridgeid != bridgeData->bd_bridgeid )
										|| 
										( configBpdu.cbpdu_portid  <= bridgeData->bd_ports[portNo]->pd_designatedport )
									)
								)	
							)
						)
					)
				)
			);
}



/*
 ***********************************************************************
 *********************OPERATION OF THE PROTOCOL*************************
 ***********************************************************************
 *															 			
 * 	These functions are used by Bridge to make operation on  		 
 *	timer expired or Bpdu received, to met the goals of 
 *	spanning tree protocol	
 *
 ***********************************************************************
 */


/* Operations on Configuration Bpdu received on a port */
void 	
stp_received_configuration_bpdu(struct stp_bridgedata *bridgeData, uint16_t portNo, struct stp_configbpdu configBpdu)
{	
	int isRoot =  False;
	isRoot = stp_root_bridge(*bridgeData);
	if (bridgeData->bd_ports[portNo]->pd_state != Disabled)
	{
		if (stp_bpdu_info_compare(bridgeData,portNo,configBpdu))
			{	
				stp_record_configuration_information(bridgeData,portNo,configBpdu);
				stp_configuration_update(bridgeData);
				stp_port_state_selection(bridgeData);
					if ( (!stp_root_bridge(*bridgeData)) && ( isRoot ) )
					{	
						stp_stop_timer(&bridgeData->bd_timerhello); 
						if ( bridgeData->bd_topologychangedetected )
							{	stp_stop_timer(&bridgeData->bd_timertopologychange);
								stp_transmit_tcn_bpdu(bridgeData);
								stp_start_timer(&bridgeData->bd_timertcn,bridgeData->bd_bridgehellotime);
							}
					}
				if ( stp_root_port(bridgeData,portNo))
					{	stp_record_configuration_timeout_values(bridgeData,configBpdu);
						stp_configuration_bpdu_generation(bridgeData);
						if ( stp_is_tcack_set(configBpdu.cbpdu_flags) ) 
							{	stp_topology_change_ack(bridgeData); }
					}
			}
		else
			{	
				if 	(stp_designated_port(bridgeData,portNo))
					{	stp_replay_to_configuration_bpdu(bridgeData,portNo); }
			}
	}
}

/*	Operations on a Topology Change Bpdu received on a port */
void	
stp_received_tcn_bpdu(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	if (bridgeData->bd_ports[portNo]->pd_state != Disabled)
		{
			stp_topology_change_detection(bridgeData);
			stp_ack_topology_change(bridgeData,portNo);
		}
}

		

/*	Operations on Hello timer expired */
static void 	
stp_hello_timer_expiry(struct stp_bridgedata *bridgeData)
{	
	stp_configuration_bpdu_generation(bridgeData);
	stp_start_timer(&bridgeData->bd_timerhello,bridgeData->bd_hellotime);
}

/*	Operations on Message Age timer expired for a port */
static void 	
stp_message_age_timer_expiry(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	int 	isRoot;
	isRoot = stp_root_bridge(*bridgeData);
	stp_become_designated_port(bridgeData,portNo);
	stp_configuration_update(bridgeData);
	stp_port_state_selection(bridgeData);
	if ( (stp_root_bridge(*bridgeData)) && (!(isRoot)) ) 
		{	bridgeData->bd_maxage = bridgeData->bd_bridgemaxage;
			bridgeData->bd_hellotime = bridgeData->bd_bridgehellotime;
			bridgeData->bd_forwarddelaytime = bridgeData->bd_bridgeforwarddelaytime;
			stp_topology_change_detection(bridgeData);
			stp_stop_timer(&bridgeData->bd_timertcn);
			stp_configuration_bpdu_generation(bridgeData);
			stp_start_timer(&bridgeData->bd_timerhello,bridgeData->bd_hellotime);
		}
}


/* Return true if some port on Bridge is designated */
static int
stp_designated_for_some_port(struct stp_bridgedata bridgeData)
	{
		uint16_t portNo;
		for(portNo=One; portNo<=bridgeData.bd_numofports; portNo++)
			{	if (bridgeData.bd_ports[portNo]->pd_designatedbridge == bridgeData.bd_bridgeid)
				{	return (True); }
			}
		return False;
	}

/*	Operations on Forward Delay Timer EXpired on a port */
static void 	
stp_forward_delay_timer_expiry(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	if ( bridgeData->bd_ports[portNo]->pd_state == Listening )
		{	
			stp_set_port_state(bridgeData->bd_ports[portNo],Learning); 
			stp_start_timer(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay,bridgeData->bd_forwarddelaytime);
		}
	else
		{	
			if ( bridgeData->bd_ports[portNo]->pd_state == Learning )
				{	
					stp_set_port_state(bridgeData->bd_ports[portNo],Forwarding);
					if (stp_designated_for_some_port(*bridgeData) && (bridgeData->bd_ports[portNo]->pd_changedetectionenabled == True))
						{	
							stp_topology_change_detection(bridgeData);
						}
				}
		}
}

/*	Operations on Topology Change Notification Timer Expired */
static void 	
stp_tcn_timer_expiry(struct stp_bridgedata *bridgeData)
{	
	stp_transmit_tcn_bpdu(bridgeData);
	stp_start_timer(&bridgeData->bd_timertcn,bridgeData->bd_bridgehellotime);
}

/*  Operations on Topology Change Timer expired */
static void 	
stp_topology_change_timer_expiry(struct stp_bridgedata *bridgeData)
{	
	bridgeData->bd_topologychangedetected = False;
	bridgeData->bd_topologychange = False;
}

/*	Operations on Hold Timer expired */
static void 	
stp_hold_timer_expiry(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	if ( bridgeData->bd_ports[portNo]->pd_configpending )
		{	stp_transmit_configuration_bpdu(bridgeData,portNo); }
}

/*  Timer implementation */
void 
stp_tick(struct stp_bridgedata *bridgeData)
{ 
	uint16_t	portNo;
	if (stp_timer_expired(&bridgeData->bd_timerhello))
		{	
			stp_hello_timer_expiry(bridgeData); 
		}
	if (stp_timer_expired(&bridgeData->bd_timertcn))
		{	
			stp_tcn_timer_expiry(bridgeData);
		}
	if (stp_timer_expired(&bridgeData->bd_timertopologychange))
		{	
			stp_topology_change_timer_expiry(bridgeData);
		}
		
	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	if (stp_timer_expired(&bridgeData->bd_ports[portNo]->pd_timermessageage))
				{	
					stp_message_age_timer_expiry(bridgeData,portNo);
				}
		}
	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	if (stp_timer_expired(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay))
				{	
					stp_forward_delay_timer_expiry(bridgeData,portNo); 
				}
			if (stp_timer_expired(&bridgeData->bd_ports[portNo]->pd_timerhold))
				{	
					stp_hold_timer_expiry(bridgeData,portNo);
				}
		}
}

/**************************************************************
 ************MANAGEMENT OF THE BRIDGE PROTOCOL ENTITY**********
 **************************************************************/

/* forward declarations */
void stp_initialize_port(struct stp_bridgedata*, uint16_t);

/* 	
 *	Inizialization process
 *	This function is used to initialize the structures, databases,
 *	and timers of the Bridge.
 */
void	
stp_initialization(struct stp_bridgedata *bridgeData)
{	
	uint16_t	portNo;

	bridgeData->bd_designatedroot = bridgeData->bd_bridgeid;
	bridgeData->bd_rootpathcost = Zero;
	bridgeData->bd_rootport = Zero;
	
	bridgeData->bd_maxage = bridgeData->bd_bridgemaxage;
	bridgeData->bd_hellotime = bridgeData->bd_bridgehellotime;
	bridgeData->bd_forwarddelaytime = bridgeData->bd_bridgeforwarddelaytime;
		
	bridgeData->bd_topologychangedetected = False;
	bridgeData->bd_topologychange = False;
	
	stp_stop_timer(&bridgeData->bd_timertopologychange);
	stp_stop_timer(&bridgeData->bd_timertcn);
	
	bridgeData->bd_ports[Zero]->pd_designatedroot = bridgeData->bd_bridgeid;
	bridgeData->bd_ports[Zero]->pd_pathcost = Zero;
	bridgeData->bd_ports[Zero]->pd_designatedcost = Zero;

	for(portNo=One; portNo <= bridgeData->bd_numofports; portNo++)
		{	
		stp_initialize_port(bridgeData,portNo);
		}
		
	stp_port_state_selection(bridgeData);
	stp_configuration_bpdu_generation(bridgeData);
	stp_start_timer(&bridgeData->bd_timerhello,bridgeData->bd_hellotime);
}

/* 	
 *	Initialization port
 *	This function is used to initialize Port structures and timers
 */
void	
stp_initialize_port(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	stp_become_designated_port(bridgeData,portNo);
	stp_set_port_state(bridgeData->bd_ports[portNo],Blocking);
	bridgeData->bd_ports[portNo]->pd_topologychangeack = False;
	bridgeData->bd_ports[portNo]->pd_configpending = False;
	
	stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timermessageage);
	stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay);
	stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timerhold);
	
	bridgeData->bd_ports[portNo]->pd_changedetectionenabled = True;
}

/* 	
 *	Enabling Port procedure
 * 	When Bridge Management operations are supported
 *	this function set the port state to blocking until
 *	operations of protocol set it to another state or leave it
 */
void 	
stp_enable_port(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	stp_initialize_port(bridgeData,portNo);
	
	stp_port_state_selection(bridgeData);
}

/* 	
 * Disaling Port procedure
 * Set of actions to do before change the port state to Disable state
 */
void 	
stp_disable_port(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	int 	isRoot = stp_root_bridge(*bridgeData);

	stp_become_designated_port(bridgeData,portNo);
	
	stp_set_port_state(bridgeData->bd_ports[portNo],Disabled);

	bridgeData->bd_ports[portNo]->pd_topologychangeack = False;
	bridgeData->bd_ports[portNo]->pd_configpending = False;
	
	stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timermessageage);
	stp_stop_timer(&bridgeData->bd_ports[portNo]->pd_timerforwarddelay);
	
	stp_configuration_update(bridgeData);
	
	stp_port_state_selection(bridgeData);
		
	if (stp_root_bridge(*bridgeData) && (!isRoot))
		{	
			bridgeData->bd_maxage = bridgeData->bd_bridgemaxage;
			bridgeData->bd_hellotime = bridgeData->bd_bridgehellotime;
			bridgeData->bd_forwarddelaytime = bridgeData->bd_bridgehellotime;
		
			stp_topology_change_detection(bridgeData);
		
			stp_stop_timer(&bridgeData->bd_timertcn);
		
			stp_configuration_bpdu_generation(bridgeData);
		
			stp_start_timer(&bridgeData->bd_timerhello,bridgeData->bd_hellotime);
		}
}

/* 	
 *	Setting Bridge Priority procedure
 * 	Used to set the most significant 16 bits of Bridge Identifier with priority p
 *	The changing of this parameter , priority, cause the reconfiguration of
 *	active topology depending on new priority assigned
 */
void 	
stp_set_bridge_priority(struct stp_bridgedata *bridgeData, uint16_t p)
{	
	uint16_t portNo;
	uint64_t newBridgeId = Zero;
	int isRoot = stp_root_bridge(*bridgeData);

	stp_set_address_bridgeid(&newBridgeId,stp_get_address_bridgeid(bridgeData->bd_bridgeid));
	stp_set_priority_bridgeid(&newBridgeId,p);
	
	for(portNo=One; portNo <= bridgeData->bd_numofports;  portNo++)
		{	if (stp_designated_port(bridgeData,portNo))
				{	bridgeData->bd_ports[portNo]->pd_designatedbridge = newBridgeId;
				}
		}
		
	bridgeData->bd_bridgeid = newBridgeId;
		
	stp_configuration_update(bridgeData);
		
	stp_port_state_selection(bridgeData);
		
	if (stp_root_bridge(*bridgeData) && (!isRoot))
		{	
			bridgeData->bd_maxage = bridgeData->bd_bridgemaxage;
			bridgeData->bd_hellotime = bridgeData->bd_bridgehellotime;
			bridgeData->bd_forwarddelaytime = bridgeData->bd_bridgeforwarddelaytime;
			
			stp_topology_change_detection(bridgeData);
			stp_stop_timer(&bridgeData->bd_timertcn);
			stp_configuration_bpdu_generation(bridgeData);
			
			stp_start_timer(&bridgeData->bd_timerhello,bridgeData->bd_hellotime);
		}
}

/* 	
 *	Setting Port Priority procedure
 *	Used to set port priority , when Bridge Management is supported
 *	and it cause the Spaning Tree reconfiguration depending on
 *	the new priority
 */
void 	
stp_set_port_priority(struct stp_bridgedata *bridgeData, uint16_t portNo, uint16_t p)
{	
	uint16_t newPortId = 0;

	stp_set_portno_portid(&newPortId,stp_get_portno_portid(bridgeData->bd_ports[portNo]->pd_portid));
	stp_set_priority_portid(&newPortId,p);

	if (stp_designated_port(bridgeData,portNo))
		{	
		bridgeData->bd_ports[portNo]->pd_designatedport = newPortId;
		}
	bridgeData->bd_ports[portNo]->pd_portid = newPortId;
	if ( (stp_designated_bridge(bridgeData,portNo)) && ( bridgeData->bd_ports[portNo]->pd_portid < bridgeData->bd_ports[portNo]->pd_designatedport) )
		{	
			stp_become_designated_port(bridgeData,portNo);
			stp_port_state_selection(bridgeData);
		}
}

/* 	
 *	Setting Path Cost procedure
 * 	To set new path cost to the link after that
 *	Spanning Tree will be reconfigured because new
 *	cost of the path to the root need to be calculated
 */
void 	
stp_set_path_cost(struct stp_bridgedata *bridgeData, uint16_t portNo, uint32_t cost)
{	
	bridgeData->bd_ports[portNo]->pd_pathcost = cost;
	stp_configuration_update(bridgeData);
	stp_port_state_selection(bridgeData);
}

/* 	
 *	Setting Port Topology Change Detection Enabled
 * 	This function set a parameter Topology Change Detecion associated to the port portNo
 *	to True in case Bridge Management activate the ability to detect Change in topology from 
 *	port "portNo"
 */
void 	
stp_enable_change_detection(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	bridgeData->bd_ports[portNo]->pd_changedetectionenabled = True;
}

/* 	
 *	Setting Port Topology Change Detection Disabled
 *	As above , but it set the parameter to False
 */
void 	
stp_disable_change_detection(struct stp_bridgedata *bridgeData, uint16_t portNo)
{	
	bridgeData->bd_ports[portNo]->pd_changedetectionenabled = False;
}
